import org.drools.util.ObjectHashMap;
import java.util.*;

public class TestMaps {
  static final int count = 100000;
  static final String[] keys = new String[count];
  static final String[] values = new String[count];
  static {
    for (int i=0; i < count; ++i) {
      keys[i] = "key" + i;
      values[i] = "value" + i;
    }
  }
  
  public static void testJUHashMap1() {
    java.util.HashMap map = new java.util.HashMap();
    
    long start = System.nanoTime();
    for (int i=0; i < count; ++i) {
      map.put(keys[i], values[i]);
    }
    long end = System.nanoTime();

    range = end-start;
    report("java.util.HashMap", "put", (end-start));
  }
  
  public static void testObjectHashMap1 () {
    ObjectHashMap map = new ObjectHashMap();

    long start = System.nanoTime();
    for (int i=0; i < count; ++i) {
      map.put(keys[i], values[i], false);
    }
    long end = System.nanoTime();
    
    report("ObjectHashMap", "put", (end-start));
  }
  
  public static void testLinearProbeMap1 () {
    LinearProbeMap map = new LinearProbeMap();

    long start = System.nanoTime();
    for (int i=0; i < count; ++i) {
      map.put(keys[i], values[i]);
    }
    long end = System.nanoTime();
    
    report("LinearProbeMap", "put", (end-start));
  }
 
  public static void testJUHashMap2() {
    java.util.HashMap map = new java.util.HashMap();
    
    for (int i=0; i < count; ++i) { map.put(keys[i], values[i]); }
    
    long start = System.nanoTime();
    for (int i=0; i < count; ++i) {
      map.get(keys[i]);
    }
    long end = System.nanoTime();
    
    range = end - start;
    report("java.util.HashMap", "get", (end-start));
  }
 
  public static void testObjectHashMap2() {
    ObjectHashMap map = new ObjectHashMap();
    
    for (int i=0; i < count; ++i) { map.put(keys[i], values[i]); }
    
    long start = System.nanoTime();
    for (int i=0; i < count; ++i) {
      map.get(keys[i]);
    }
    
    long end = System.nanoTime();
    report("ObjectHashMap", "get", (end-start));
  }
 
  public static void testLinearProbeMap2() {
    LinearProbeMap map = new LinearProbeMap();
    
    for (int i=0; i < count; ++i) { map.put(keys[i], values[i]); }
    
    long start = System.nanoTime();
    for (int i=0; i < count; ++i) {
      map.get(keys[i]);
    }
    
    long end = System.nanoTime();
    report("LinearProbeMap", "get", (end-start));
  }

  public static void testJUHashMap3() {
    java.util.HashMap map = new java.util.HashMap();
    
    for (int i=0; i < count; ++i) { map.put(keys[i], values[i]); }
    
    long start = System.nanoTime();
    java.util.Iterator itr = map.values().iterator();
    while ( itr.hasNext()) {
      itr.next().hashCode();
    }
    long end = System.nanoTime();
    
    range = end-start;
    report("java.util.HashMap", "iterate", (end-start));
  }
  
  public static void testObjectHashMap3() {
    ObjectHashMap map = new ObjectHashMap();
    
    for (int i=0; i < count; ++i) { map.put(keys[i], values[i]); }
    
    long start = System.nanoTime();
    org.drools.util.Iterator itr = map.iterator();
    Object val = null;
    while ( (val = itr.next()) != null) {
      val.hashCode();
    }
    long end = System.nanoTime();
    
    report("ObjectHashMap", "iterate", (end-start));
  }

  public static void testLinearProbeMap3() {
    LinearProbeMap map = new LinearProbeMap();
    
    for (int i=0; i < count; ++i) { map.put(keys[i], values[i]); }
    
    long start = System.nanoTime();
    java.util.Iterator<Object> itr = map.iterator();
    while ( itr.hasNext()) {
      itr.next().hashCode();
    }
    long end = System.nanoTime();
    
    report("LinearProbeMap", "iterate", (end-start));
  }

  static long    range = 1;
  static TreeMap<String, long[]> results = new TreeMap();
  static TreeMap<String, long[]> ranges = new TreeMap();
  static int     runs;
  static int     run;
  
  public static void main (String...args) {
    runs = args.length > 0 ? Integer.parseInt(args[0]) : 8;
    
    for (run = 0; run < runs; ++run) {
      testJUHashMap1();
      testObjectHashMap1();
      testLinearProbeMap1();
      System.out.println();
      testJUHashMap2();
      testObjectHashMap2();
      testLinearProbeMap2();
      System.out.println();
      testJUHashMap3();
      testObjectHashMap3();
      testLinearProbeMap3();
      System.out.println();
      System.out.println();
    }
    
    reportAverages();
  }
  
  static final String spaces = "                                        ";
  
  static void report (String className, String test, long result) {
    String prefix = test + ' ' + className;
    System.out.println(prefix + spaces.substring(prefix.length()) + formatNanos(result));

    long[] timings = (long[])results.get(prefix);
    
    if (timings == null) {
      timings = new long[runs];
      results.put(prefix, timings);
    }
    
    timings[run] = result;

    // store ranges
    timings = (long[])ranges.get(test);
    
    if (timings == null) {
      timings = new long[runs];
      ranges.put(test, timings);
    }
    
    if (timings[run] == 0) timings[run] += result;
    
    System.gc();
  }
  
  static void reportAverages () {
    if (runs > 7) {
      int outliers = runs / 8;
      System.out.println("Mean of " + runs + " runs excluding best and worst " + outliers + ":");
      for (Map.Entry<String, long[]> e : results.entrySet()) {
        final String prefix = e.getKey();
        final long result;
        {
          long[] timings = e.getValue();
          Arrays.sort(timings);
          long total = 0;
          for (int i = outliers; i < timings.length - outliers; ++i) total += timings[i];
          result = (total / (timings.length - 2*outliers));
        }
        
        // get first timings for this test
        int    index = prefix.indexOf(' ');
        String test = prefix.substring(0, index);
        
        {
          long[] timings = ranges.get(test);
          Arrays.sort(timings);
          long total = 0;
          for (int i = outliers; i < timings.length - outliers; ++i) total += timings[i];
          range = (total / (timings.length - 2*outliers));
        }
        
        System.out.println(prefix + spaces.substring(prefix.length()) + formatNanos(result));
      }
    }
  }
  
  static String stars = "*******************|*******************|*******************|";
  static String formatNanos (long nanos) {
    int bar = (int)((range * 20.0) / nanos);
    String millis = "      " + (nanos / 1000000L);
    String micros = "00" + ((nanos / 1000L) % 1000L);
    
    if (bar > stars.length()) {
      return millis.substring(millis.length()-6) + '.' + micros.substring(micros.length()-3) + " ms " + stars + '+';
    } else {
      return millis.substring(millis.length()-6) + '.' + micros.substring(micros.length()-3) + " ms " + stars.substring(0,bar);
    }
  }
}